import math
from decimal import Decimal

from .round_half_up import round_half_up
from .round_up import round_up

__all__ = ["round_half_away_from_zero"]


def round_half_away_from_zero(n: float | Decimal, places: int | None = None):
    if isinstance(n, Decimal):
        places = Decimal("1" if not places else f"1.{str(0) * places}")
        return round_up(n, places=places)
    rounded_abs = round_half_up(abs(n), places)
    return math.copysign(rounded_abs, n)



MW_CREATININE = 113.12  # g/mol

# in g/mol
molecular_weights = dict(
    albumin=66000,
    chol=386.65,
    creatinine=113.12,
    glucose=180.16,
    haemoglobin=64500,
    tbil=584.6,
    trig=885,
    urea=60.06,
)



from __future__ import annotations

from clinicedc_constants import FEMALE, MALE, MICROMOLES_PER_LITER, MILLIGRAMS_PER_DECILITER

from ..constants import MW_CREATININE
from ..exceptions import EgfrCalculatorError

__all__ = ["BaseEgfr"]


class BaseEgfr:
    male = MALE
    female = FEMALE
    micromoles_per_liter = MICROMOLES_PER_LITER
    milligrams_per_deciliter = MILLIGRAMS_PER_DECILITER
    min_age_in_years = 18.0
    max_age_in_years = 120.0
    mw_creatinine = MW_CREATININE  # g/mol

    def __init__(
        self,
        gender: str | None = None,
        age_in_years: int | float | None = None,
        creatinine_value: float | int | None = None,
        creatinine_units: str | None = None,
        **kwargs,  # noqa
    ):
        self.creatinine_value = creatinine_value
        self.creatinine_units = creatinine_units

        if not gender or gender not in [self.male, self.female]:
            raise EgfrCalculatorError(
                f"Invalid gender. Expected one of {self.male}, {self.female}. Got {gender}."
            )
        self.gender = gender

        if not (self.min_age_in_years <= (age_in_years or 0.0) < self.max_age_in_years):
            raise EgfrCalculatorError(
                f"Invalid age. See {self.__class__.__name__}. Got {age_in_years}"
            )
        self.age_in_years = float(age_in_years) if age_in_years else None



from clinicedc_constants import (
    GRAMS_PER_LITER,
    MICROMOLES_PER_LITER,
    MILLIGRAMS_PER_DECILITER,
    MILLIMOLES_PER_LITER,
)

from ..exceptions import ConversionNotHandled

__all__ = ["micromoles_per_liter_to"]


def micromoles_per_liter_to(
    *, value: float | int, units_to: str, mw: float | None = None
) -> dict:
    if units_to == MICROMOLES_PER_LITER:
        return {MICROMOLES_PER_LITER: float(value)}
    if units_to == MILLIMOLES_PER_LITER:
        return {MILLIMOLES_PER_LITER: float(value) / 1000.00}
    if units_to == GRAMS_PER_LITER:
        if mw is None:
            raise ConversionNotHandled("Molecular weight may not be None.")
        return {GRAMS_PER_LITER: (float(value) * mw) / 100.00}
    if units_to == MILLIGRAMS_PER_DECILITER:
        if mw is None:
            raise ConversionNotHandled("Molecular weight may not be None.")
        return {MILLIGRAMS_PER_DECILITER: (float(value) * mw) / 10000.00}
    raise ConversionNotHandled(f"Conversion not found. Tried umol/L to {units_to}. ")



from .egfr_ckd_epi import EgfrCkdEpi
from .egfr_cockcroft_gault import EgfrCockcroftGault
from .egfr_percent_change import egfr_percent_change

__all__ = ["EgfrCkdEpi", "EgfrCockcroftGault", "egfr_percent_change"]



from __future__ import annotations

import decimal
from decimal import Decimal

__all__ = ["round_up"]


def round_up(value: Decimal, places: Decimal | None = None):
    places = places or Decimal("1.0000")
    if value > Decimal("0.0000000000"):
        decimal.getcontext().rounding = decimal.ROUND_HALF_UP
    else:
        decimal.getcontext().rounding = decimal.ROUND_HALF_DOWN
    return value.quantize(places)



from __future__ import annotations

__all__ = ["convert_units"]

from .units_converter import UnitsConverter


def convert_units(
    *,
    label: str | None = None,
    value: int | float | None = None,
    units_from: str | None = None,
    units_to: str | None = None,
    places: int | None = None,
    mw: float | None = None,
) -> int | float:
    return UnitsConverter(
        label=label,
        value=value,
        units_from=units_from,
        units_to=units_to,
        places=places,
        mw=mw,
    ).converted_value



from .convert_units import convert_units, micromoles_per_liter_to, milligrams_per_deciliter_to
from .egfr_calculators import EgfrCkdEpi, EgfrCockcroftGault, egfr_percent_change
from .exceptions import ConversionNotHandled, EgfrCalculatorError
from .round_up import round_half_away_from_zero, round_half_up, round_up

__all__ = [
    "ConversionNotHandled",
    "EgfrCalculatorError",
    "EgfrCkdEpi",
    "EgfrCockcroftGault",
    "convert_units",
    "egfr_percent_change",
    "micromoles_per_liter_to",
    "milligrams_per_deciliter_to",
    "round_half_away_from_zero",
    "round_half_up",
    "round_up",
]



from .round_half_away_from_zero import round_half_away_from_zero
from .round_half_up import round_half_up
from .round_up import round_up

__all__ = ["round_half_away_from_zero", "round_half_up", "round_up"]



from .convert_units import convert_units
from .micromoles_per_liter_to import micromoles_per_liter_to
from .milligrams_per_deciliter_to import milligrams_per_deciliter_to

__all__ = [
    "convert_units",
    "micromoles_per_liter_to",
    "milligrams_per_deciliter_to",
]



from __future__ import annotations

from clinicedc_constants import BLACK

from clinicedc_utils import convert_units

from .base_egfr import BaseEgfr, EgfrCalculatorError

# TODO: https://www.rcpa.edu.au/Manuals/RCPA-Manual/
#  Pathology-Tests/C/Creatinine-clearance-Cockcroft-and-Gault

__all__ = ["EgfrCkdEpi"]


class EgfrCkdEpi(BaseEgfr):
    """Reference https://nephron.com/epi_equation

    CKD-EPI Creatinine equation (2009)

    Levey AS, Stevens LA, et al. A New Equation to Estimate Glomerular
    Filtration Rate. Ann Intern Med. 2009; 150:604-612.
    """

    black = BLACK

    def __init__(self, ethnicity: str, **kwargs):
        super().__init__(**kwargs)
        self.ethnicity = ethnicity
        if self.creatinine_units == self.micromoles_per_liter:
            self.creatinine_value = convert_units(
                label="creatinine",
                value=float(self.creatinine_value),
                units_from=self.creatinine_units,
                units_to=self.milligrams_per_deciliter,
                mw=self.mw_creatinine,
            )

    @property
    def value(self) -> float | None:
        if self.gender and self.age_in_years and self.ethnicity and self.creatinine_value:
            return float(
                141.000
                * (min(self.creatinine_value / self.kappa, 1.000) ** self.alpha)
                * (max(self.creatinine_value / self.kappa, 1.000) ** -1.209)
                * self.age_factor
                * self.gender_factor
                * self.ethnicity_factor
            )
        opts = dict(
            gender=self.gender,
            age_in_years=self.age_in_years,
            ethnicity=self.ethnicity,
            creatinine_value=self.creatinine_value,
        )
        raise EgfrCalculatorError(f"Unable to calculate. Insufficient information. Got {opts}")

    @property
    def alpha(self) -> float:
        return float(-0.329 if self.gender == self.female else -0.411)

    @property
    def kappa(self) -> float:
        return float(0.7 if self.gender == self.female else 0.9)

    @property
    def ethnicity_factor(self) -> float:
        return float(1.159 if self.ethnicity == self.black else 1.000)

    @property
    def gender_factor(self) -> float:
        return float(1.018 if self.gender == self.female else 1.000)

    @property
    def age_factor(self) -> float:
        return float(0.993**self.age_in_years)



from clinicedc_constants import (
    GRAMS_PER_LITER,
    MICROMOLES_PER_LITER,
    MILLIGRAMS_PER_DECILITER,
    MILLIGRAMS_PER_LITER,
    MILLIMOLES_PER_LITER,
)

from ..exceptions import ConversionNotHandled

__all__ = ["milligrams_per_deciliter_to"]


def milligrams_per_deciliter_to(
    *, value: float | int, units_to: str, mw: float | None = None
) -> dict[str, float]:
    if units_to == MILLIGRAMS_PER_DECILITER:
        return {MILLIGRAMS_PER_DECILITER: float(value)}
    if units_to == MILLIMOLES_PER_LITER:
        if mw is None:
            raise ConversionNotHandled("Molecular weight may not be None.")
        return {MILLIMOLES_PER_LITER: (float(value) * 10.00) / mw}
    if units_to == MICROMOLES_PER_LITER:
        if mw is None:
            raise ConversionNotHandled("Molecular weight may not be None.")
        return {MICROMOLES_PER_LITER: (float(value) * 10000.00) / mw}
    if units_to == MILLIGRAMS_PER_LITER:
        return {MILLIGRAMS_PER_LITER: float(value) * 10.00}
    if units_to == GRAMS_PER_LITER:
        return {GRAMS_PER_LITER: float(value) / 100.00}
    raise ConversionNotHandled(f"Conversion not found. Tried mg/dL to {units_to}. ")



from clinicedc_constants import (
    GRAMS_PER_LITER,
    MICROMOLES_PER_LITER,
    MILLIGRAMS_PER_DECILITER,
    MILLIMOLES_PER_LITER,
)

from ..constants import molecular_weights
from ..exceptions import ConversionNotHandled
from ..round_up import round_half_away_from_zero
from .micromoles_per_liter_to import micromoles_per_liter_to
from .milligrams_per_deciliter_to import milligrams_per_deciliter_to

__all__ = ["UnitsConverter"]


class UnitsConverter:
    def __init__(
        self,
        *,
        label: str,
        value: int | float,
        units_from: str,
        units_to: str,
        places: int | None = None,
        mw: float | None = None,
    ):
        self.label = label
        self.value = value
        self.units_from = units_from
        self.units_to = units_to
        self.places = places or 4
        self._mw = mw
        if label is None:
            raise ValueError("label is required. See convert_units.")
        if value is not None and units_from and units_to and units_from != units_to:
            self.converted_value = self.get_converted_value()
        elif units_from == units_to:
            self.converted_value = value
        else:
            raise ConversionNotHandled(
                f"Conversion not handled. Tried {label} from {units_from} to {units_to}."
            )

    @property
    def mw(self) -> float | int:
        if self._mw is None:
            try:
                self._mw = molecular_weights[self.label]
            except KeyError as e:
                raise ConversionNotHandled(
                    f"Molecular weight may not be None. Got {self.label}."
                ) from e
        return self._mw

    def round_up(self, converted_value):
        try:
            converted_value = round_half_away_from_zero(converted_value, self.places)
        except TypeError as e:
            raise ConversionNotHandled(
                f"Conversion not handled. Tried {self.label} from {self.units_from} "
                f"to {self.units_to}. "
                f"Got {e} when rounding {converted_value} to {self.places} places."
            ) from e
        return converted_value

    def from_milligrams_per_deciliter(self) -> float | int:
        if self.units_to != MILLIGRAMS_PER_DECILITER:
            return milligrams_per_deciliter_to(
                value=float(self.value), units_to=self.units_to, mw=self.mw
            )[self.units_to]
        return self.value

    def from_grams_per_liter(self) -> float | int:
        if self.units_to != GRAMS_PER_LITER:
            return milligrams_per_deciliter_to(
                value=float(self.value) * 100.00,
                units_to=self.units_to,
                mw=self.mw,
            )[self.units_to]
        return self.value

    def from_millimoles_per_liter(self) -> float | int:
        if self.units_to != MILLIMOLES_PER_LITER:
            return micromoles_per_liter_to(
                value=self.value / 1000, units_to=self.units_to, mw=self.mw
            )[self.units_to]
        return self.value

    def from_micromoles_per_liter(self):
        if self.units_to != MICROMOLES_PER_LITER:
            return micromoles_per_liter_to(
                value=self.value, units_to=self.units_to, mw=self.mw
            )[self.units_to]
        return self.value

    def get_converted_value(self) -> int | float:
        converted_value = None
        if self.units_from == MILLIGRAMS_PER_DECILITER:
            converted_value = self.from_milligrams_per_deciliter()
        elif self.units_from == GRAMS_PER_LITER:
            converted_value = self.from_grams_per_liter()
        elif self.units_from == MILLIMOLES_PER_LITER:
            converted_value = self.from_millimoles_per_liter()
        elif self.units_from == MICROMOLES_PER_LITER:
            converted_value = self.from_micromoles_per_liter()
        if not converted_value:
            raise ConversionNotHandled(
                f"Conversion not handled. Tried {self.label} from "
                f"{self.units_from} to {self.units_to}."
            )
        return self.round_up(converted_value)
